#! /usr/bin/env python3
#
# test via curl:
#
#   curl --silent -H "Content-Type: application/json" -X POST -d '{"command":"messages", "limit":10}' 127.0.0.1:5000/api/message
#
# Copyright (c) 2023 ZKM | Hertz-Lab
# Dan Wilcox <dan.wilcox@zkm.de>
#
# BSD Simplified License.
# For information on usage and redistribution, and for a DISCLAIMER OF ALL
# WARRANTIES, see the file, "LICENSE.txt," in this distribution.
#
# This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
# Museum“ generously funded by the German Federal Cultural Foundation.
#
# refs:
#   * https://pythonbasics.org/flask-sqlite/
#   * https://flask.palletsprojects.com/en/2.3.x/tutorial/database/
#   * https://stackabuse.com/how-to-get-and-parse-http-post-body-in-flask-json-and-form-data/

# FIXME: hack to find shared module
import os, sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))

import json, argparse
from numbers import Number

from flask import Flask, render_template, request, g
import things
import query

##### parser

parser = argparse.ArgumentParser(description="MQTT message database API server")
parser.add_argument(
    "--db", action="store", dest="db",
    default="messages.db", help="SQLite database path")
parser.add_argument(
    "-v", "--verbose", action="store_true", dest="verbose",
    help="enable verbose printing")

##### database

# get/create database instance
def db_get():
    if "db" not in g:
        db = things.Database(verbose=args.verbose)
        if db.open(args.db): g.db = db
    return g.db

# close current database instance
def db_close(e=None):
    db = None
    if "db" in g: db = g.pop("db")
    if db: db.close()

##### api

# returns True if limit key exists in data dict and is a valid number
def api_valid_limit(limit):
    return isinstance(limit, Number)

# returns True if order key exists in data dict and has a valid value
def api_valid_order(order):
    return isinstance(order, str) and (order == "desc" or order == "asc")

# format failure json response 
def api_resp_failure(message):
   return json.dumps({"status": 1, "msg": message})

# format success json response
def api_resp_success(data, pretty=False):
    d = {"status": 0, "msg": "success", "data": data}
    if pretty:
        return json.dumps(d, indent=2, separators=(",", ": "))
    else:
        return json.dumps(d, separators=(",", ":"))

##### flask

app = Flask(__name__)

@app.route("/")
def home():
    return render_template("home.html")

@app.route("/count")
def count_messages():
    count = query.count_messages(db_get())
    return render_template("count.html", count=count)

@app.route("/topics")
def select_topics():
    rows = query.select_topics(db_get())
    return render_template("topics.html", rows=rows)

@app.route("/messages")
def select_messages():
    rows = query.select_messages(db_get(), limit=100)
    return render_template("messages.html", desc="last 100 messages", rows=rows)

@app.route("/current")
def select_current():
    rows = query.select_current_messages(db_get())
    return render_template("messages.html", desc="current message for each topic", rows=rows)

# message access json api
@app.post("/api/message")
def api_messages():
    if request.headers.get("Content-Type") != "application/json":
        return "Content-Type not supported"

    # check data
    data = request.json
    if data is None or not isinstance(data, dict):
        return api_resp_failure("invalid request")

    # check command
    if not "command" in data:
        return api_resp_failure("missing command")
    command = data["command"]
    pretty = False if not "pretty" in data else data["pretty"]

    # command handling
    if command == "count":
        count = query.count_messages(db_get())
        return api_resp_success(count, pretty=pretty)

    elif command == "topics":
        rows = query.count_messages(db_get())

    elif command == "current":
        rows = query.select_current_messages(db_get())

    # {"command": "messages", "limit": 100, "order": "desc"}
    elif command == "messages":
        limit = 1000 if not "limit" in data else data["limit"]
        if not api_valid_limit(limit):
            return api_resp_failure("limit invalid")
        order = None
        if "order" in data:
            order = data["order"]
            if not api_valid_order(order):
                return api_resp_failure("order invalid")
        rows = query.select_messages(db_get(), limit=limit, order=order)

    # {"command": "match", "topics": ["%person%"], "limit": 100, "order": "desc"}
    elif command == "match":
        topics = None if not "topics" in data else data["topics"]
        if not isinstance(topics, list):
            return api_resp_failure("topics array missing or invalid")
        limit = 0 if not "limit" in data else data["limit"]
        if not api_valid_limit(limit):
            return api_resp_failure("limit invalid")
        order = None
        if "order" in data:
            order = data["order"]
            if not api_valid_order(order):
                return api_resp_failure("order invalid")
        rows = query.match_topics(db_get(), topics, limit=limit, order=order)

    else:
        return api_resp_failure("unknown command")

    return api_resp_success(rows, pretty=pretty)

##### main

# parse args
args = parser.parse_args()

if __name__ == '__main__':
    app.teardown_appcontext(db_close)
    app.run()
