#! /bin/sh
#
# export from an Sqlite database file to stdout or csv
#
# note: there are no security checks or SQL sanitization as this is meant to be
#       run by an admin with access to the database file itself
#
# Copyright (c) 2023 ZKM | Hertz-Lab
# Dan Wilcox <dan.wilcox@zkm.de>
#
# BSD Simplified License.
# For information on usage and redistribution, and for a DISCLAIMER OF ALL
# WARRANTIES, see the file, "LICENSE.txt," in this distribution.
#
# This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
# Museum“ generously funded by the German Federal Cultural Foundation.
#
# refs:
#  * https://sqlite.org/cli.html#export_to_csv
#  * https://www.w3schools.com/sql/default.asp

##### variables

SQLITE=sqlite3
DB=messages.db
CSV=
TIMEFMT=", 'localtime'"
SEP=", "
LIMIT=
ORDER=

##### functions

# run a db query, outputs to stdout or $CSV if set
# $1: SQL query string, ie. "SELECT * FROM table;"
query() {
	if [ "$CSV" != "" ] ; then
echo "
.headers on
.mode csv
.once $CSV
$1" | $SQLITE $DB
	else
echo "
.separator '$SEP'
$1" | $SQLITE $DB
	fi
}

# check if an argument is set, print error if not and return
check_arg() {
	if [ "$2" == "" ] ; then
		echo "$1 required"
		return 1
	fi
	return 0
}

##### arguments

CMDHELP="commands: count, topics, current, messages, match, query"
HELP="Usage: export.sh [OPTIONS] [DB] CMD [ARG...]

  export MQTT messages from an SQLite db file to stdout or CSV

Options:
  --db PATH     input database path (default: $DB)
  --csv PATH    output CSV file
  --utc         format timestamps to UTC (default: local)
  --sep SEP     set stdout col separator (default: \", \")
  --limit N     limit max num of messages
  --order STR   set output order: "desc" or "asc" (default)
  -h,--help     display this help message
   
Arguments:
  CMD           command to run

Commands:
  count:        overall message count
  topics:       unique topics and message count
  current:      latest message for each topic
  messages:     all messages
  match STR...: topics matching given search STR(s) using
                SQL _ (single) and % (multi) char wildcards
                ex. test/temperature, %person%, etc
  query STR:    run SQL query string (there be dragons...)
"

while [ "$1" != "" ] ; do
  case $1 in
    --db) shift 1 ; check_arg "db PATH" $1 || exit 1 ; DB=$1 ;;
    --csv) shift 1 ; check_arg "csv PATH" $1 || exit 1 ; CSV=$1 ;;
		--utc) TIMEFMT="" ;;
		--sep) shift 1 ; check_arg "stdout SEP" $1 || exit 1 ; SEP=$1 ;;
		--limit) shift 1 ; check_arg "limit N" $1 || exit 1 ; LIMIT=$1 ;;
		--order) shift 1 ; check_arg "order STR" $1 || exit 1 ; ORDER=$1 ;;
    -h|--help) echo "$HELP" ; exit 0 ;;
    *) break ;;
  esac
  shift 1
done
if [ "$1" == "" ] ; then
	echo "$CMDHELP"
	exit 1
fi
CMD=$1
shift 1

##### main

TABLE=messages
TIME="datetime(time, 'unixepoch'$TIMEFMT) as time"
MAXTIME="datetime(MAX(time), 'unixepoch'$TIMEFMT) as time"

# command handling
case $CMD in

	count)
		query "SELECT COUNT(*) as 'count' FROM $TABLE;"
		;;

	topics)
		query "SELECT topic, COUNT(topic) as 'count' FROM $TABLE GROUP BY topic;"
		;;

	current)
		query "SELECT $MAXTIME, topic, message FROM $TABLE GROUP BY topic;"
		;;

	messages)
		q="SELECT $TIME, topic, message FROM $TABLE"
		if [ "$ORDER" != "" ] ; then q="$q ORDER BY id $ORDER" ; fi
		if [ "$LIMIT" != "" ] ; then q="$q LIMIT $LIMIT" ; fi
		query "${q};"
		;;

	match)
		check_arg "match STR" $1 || exit 1
		q="SELECT $TIME, topic, message FROM $TABLE WHERE topic LIKE '$1'"
		shift 1
		while [ "$1" != "" ] ; do
			q="$q OR topic LIKE '$1'"
			shift 1
		done
		if [ "$ORDER" != "" ] ; then q="$q ORDER BY id $ORDER" ; fi
		if [ "$LIMIT" != "" ] ; then q="$q LIMIT $LIMIT" ; fi
		query "${q};"
		;;

	query)
		query "$1"
		;;

	*) echo "$CMDHELP" ; echo "unknown command: $CMD" ; exit 1 ;;
esac
