#! /usr/bin/env python3
#
# Copyright (c) 2023 ZKM | Hertz-Lab
# Dan Wilcox <dan.wilcox@zkm.de>
#
# BSD Simplified License.
# For information on usage and redistribution, and for a DISCLAIMER OF ALL
# WARRANTIES, see the file, "LICENSE.txt," in this distribution.
#
# This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
# Museum“ generously funded by the German Federal Cultural Foundation.

# FIXME: hack to find shared module
import os, sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))

import threading, time
from queue import Queue

import things

# MQTT message logger, writes to db in background thread
class Logger:

    # init logger with sqlite database path
    def __init__(self, path, drop=False, verbose=False):
        self.path = path       # sqlite db path
        self.drop = drop       # drop table on start? 
        self.thread = None     # background thread
        self.run = True        # run control
        self.queue = Queue()   # message-passing queue
        self.pmessages = {}    # previous messages by topic key
        self.verbose = verbose # print debug info?

    # start logging
    def start(self):
        self.run = True
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    # stop logging
    def stop(self):
        self.run = False
        if self.thread:
            self.thread.join()
            self.thread = None
            self.queue.queue.clear()
            self.pmessages.clear()

    # log an MQTT message
    def log(self, topic, message):
        if not self._changed(topic, message): return
        row = [int(time.time()), topic, message]
        self.queue.put(row)
        if self.verbose: print(f"Logger: log {topic} {message}")

    # returns True if the thread is currently running
    def is_running(self):
        return self.thread and self.thread.is_alive()

    # thread run function
    def _run(self):
        if self.verbose:
            print("Logger: starting")
        db = things.Database(verbose=self.verbose)
        if not db.open(self.path): return
        if self.drop: db.drop(things.schema.TABLE)
        db.create(things.schema.TABLE, things.schema.FIELDS)
        while self.run:
            while not self.queue.empty():
                row = self.queue.get()
                if row is None: continue
                db.query(
                    "INSERT INTO messages (time, topic, message) VALUES (?, ?, ?)",
                    row, commit=True)
            time.sleep(0.01)
        db.close()
        print("Logger: stopped")

    # TODO: limit stored message size?
    # returns true if topic is new or has changed since previous message
    def _changed(self, topic, message):
        if topic in self.pmessages:
            if self.pmessages[topic] == message:
                return False
        self.pmessages[topic] = message
        return True

# commandline test
if __name__ == '__main__':
    import sys, os
    logger = Logger("test.db", drop=True, verbose=True)
    logger.start()
    if not logger.is_running():
        sys.exit(1)
    for i in range(10):
        logger.log("test", "message {}".format(i+1))
    logger.log("test", str.format("message 10")) # check dup of prev message
    time.sleep(1)
    logger.stop()
    os.system(f"echo 'SELECT * FROM {things.schema.TABLE};' | sqlite3 test.db")
