#! /usr/bin/env python3
#
# Copyright (c) 2023 ZKM | Hertz-Lab
# Dan Wilcox <dan.wilcox@zkm.de>
#
# BSD Simplified License.
# For information on usage and redistribution, and for a DISCLAIMER OF ALL
# WARRANTIES, see the file, "LICENSE.txt," in this distribution.
#
# This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
# Museum“ generously funded by the German Federal Cultural Foundation.
#
# refs:
#  * http://www.steves-internet-guide.com/logging-mqtt-sensor-data-to-sql-database-with-python/

# FIXME: hack to find shared module
import os, sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))

import time, argparse
from logger import Logger

mqtt_host = "localhost"
mqtt_port = 1883
mqtt_topics = []
mqtt_user = ""
mqtt_pass = ""
mqtt_pmessage = {}
db_path = "messages.db"

verbose = True

##### parser

parser = argparse.ArgumentParser(description="MQTT SQLite db logger")
parser.add_argument(
    "host", type=str, nargs="?", metavar="HOST",
    default="", help="MQTT server host name or IP address, optionally append port HOST:PORT")
parser.add_argument(
    "topic", type=str, nargs="+", metavar="TOPIC",
    default="", help="MQTT topic(s) to log, can be multiple or include wildcards")
parser.add_argument(
    "--user", action="store", dest="user",
    default="", help="MQTT server user name")
parser.add_argument(
    "--pass", action="store", dest="password",
    default="", help="MQTT server user password")
parser.add_argument(
    "--db", action="store", dest="db",
    default=db_path, help="SQLite database path")
parser.add_argument(
    "-v", "--verbose", action="store_true", dest="verbose",
    help="enable verbose printing")

##### mqtt

import paho.mqtt.client as mqtt

# TODO: try to reconnect?
def mqtt_on_connect(client, userdata, flags, rc):
    if rc == 0:
        if verbose: print(f"mqtt: connected")
    else: print(f"mqtt: {mqtt.error_string(rc)}")
    # stop (re)connecting if refused, ie. bad authentication
    if rc == mqtt.MQTT_ERR_CONN_REFUSED:
        client.disconnect()

# TODO: stop logging on disconnect? try to reconnect?
def mqtt_on_disconnect(client, userdata, rc):
    if rc == 0:
        if verbose: print(f"mqtt: disconnected")
    else: print(f"mqtt: {mqtt.error_string(rc)}")

# def mqtt_on_subscribe(client, userdata, mid, granted_qos):
#     pass

def mqtt_on_message(client, userdata, msg):
    try:
        payload = str(msg.payload.decode("utf-8", "ignore"))
    except Exception as exc:
        print(f"mqtt: could not parse payload: {exc}")
        return
    logger.log(msg.topic, payload)

##### main

# parse args
args = parser.parse_args()
if args.host == "" or args.topic == "":
        print("MQTT HOST & TOPIC(s) required", file=sys.stderr)
        sys.exit(1)
if ":" in args.host: # parse optional HOST:PORT
    s = args.host.split(":")
    try: # strip last component
        mqtt_port = int(s[-1])
    except:
        print(f"invalid port {s[-1]} in MQTT HOST {args.host}", file=sys.stderr)
        sys.exit(1)
    # join everything else, in case of domain ie. http://
    mqtt_host = ":".join(s[:-1])
else:
    mqtt_host = args.host
if len(args.topic) > 0:
    mqtt_topics = args.topic
if args.user != "":
    if args.password == "":
        print(f"password required for MQTT server user {args.user}", file=sys.stderr)
        sys.exit(1)
    else:
        mqtt_user = args.user
        mqtt_pass = args.password
db_path = args.db
verbose = args.verbose
args = None

# db logger
logger = Logger(db_path, verbose=verbose)
logger.start()
if not logger.is_running():
    print(f"could not start logger", file=sys.stderr)
    sys.exit(1)

# mqtt client
client = mqtt.Client()
client.on_connect = mqtt_on_connect
client.on_disconnect = mqtt_on_disconnect
#client.on_subscribe = mqtt_on_subscribe
client.on_message = mqtt_on_message

# loop
try:
    if mqtt_user:
        client.username_pw_set(username=mqtt_user, password=mqtt_pass)
    client.connect(mqtt_host, mqtt_port, 60)
    for topic in mqtt_topics:
        client.subscribe(topic)
    client.loop_forever()
except KeyboardInterrupt:
    pass
except Exception as exc:
    print(exc, file=sys.stderr)

# cleanup
time.sleep(0.1)
logger.stop()
