# Copyright (c) 2023 ZKM | Hertz-Lab
# Dan Wilcox <dan.wilcox@zkm.de>
#
# BSD Simplified License.
# For information on usage and redistribution, and for a DISCLAIMER OF ALL
# WARRANTIES, see the file, "LICENSE.txt," in this distribution.
#
# This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
# Museum“ generously funded by the German Federal Cultural Foundation.

import sys, sqlite3

# sqlite3 database wrapper
class Database():

	def __init__(self, verbose=False):
		self.con = None # sqlite connection
		self.cur = None # sqlite cursor
		self.verbose = verbose # print extra debug info?

	# open sqlite database at path, creates if file does not exist
	# returns True on success
	def open(self, path):
		try:
			self.con = sqlite3.connect(path, isolation_level=None)
			if self.verbose:
				print(f"Database: sqlite {sqlite3.version}")
				print(f"Database: opened {path}")
		except sqlite3.Error as exc:
			print(f"Database: could not open {path}: {exc}", file=sys.stderr)
			self.con = None
			return False
		self.con.execute("pragma foreign_keys = on")
		self.con.execute("pragma journal_mode = wal")
		self.con.commit()
		self.cur = self.con.cursor()
		return True

	# close database
	def close(self):
		if self.con == None: return
		self.cur = None
		self.con.close()
		if self.verbose:
			print(f"Database: closed")

	# create table
	# table: table name
	# fields: dictionary of field name keys and type string values
	# example:
	#     db.create("logs", {"id": "INTEGER PRIMARY KEY AUTOINCREMENT",
	#                        "time": "INTEGER",
	#                        "topic": "TEXT",
	#                        "message": "TEXT"}
	# returns True on success
	def create(self, table, fields):
		q = "CREATE TABLE IF NOT EXISTS " + table + " ("
		for key,value in fields.items():
			q = q + key + " " + value + ","
		q = q[:-1]
		q = q + ");"
		try:
			self.cur.execute(q)
			if self.verbose:
				print(f"Database: created table {table}")
			return True
		except sqlite3.Error as exc:
			print(f"Database: {exc}", file=sys.stderr)
		return False

	# drop a table
	# table: table name
	# returns True on success
	def drop(self, table):
		try:
			self.cur.execute("DROP TABLE " + table + ";")
			if self.verbose:
				print(f"Database: dropped table {table}")
			return True
		except sqlite3.Error as exc:
			print(f"Database: {exc}", file=sys.stderr)
		return False

	# perform an SQL query with optional args
	# example:
	#    db.query("INSERT INTO logs (time, topic, message) VALUES (?, ?, ?)",
	#             [timestamp, topic, message], commit=True)
	# note: set commit=True on changes (INSERT, UPDATE, SET) to ensure
	#       propogation to other connections, not required for queries like SELECT
	# returns rows array on success or None on failure,
	#         rows may be empty depending upon the query
	def query(self, query, args=(), commit=False):
		try:
			self.cur.execute(query, args)
			if commit:
				self.con.commit()
			return self.cur.fetchall()
		except sqlite3.Error as exc:
			print(f"Database: {exc}", file=sys.stderr)
		return None

# commandline test
if __name__ == '__main__':
	import time
	db = Database(verbose=True)
	if not db.open("test.db"):
		sys.exit(1)
	fields = {
		"id": "INTEGER PRIMARY KEY AUTOINCREMENT",
		"time": "INTEGER",
		"topic": "TEXT",
		"message": "TEXT"
	}
	db.drop("test")
	db.create("test", fields)
	db.query("INSERT INTO test (time, topic, message) VALUES (?, ?, ?)",
		[int(time.time()), "foo", "hello world"], commit=True)
	rows = db.query("SELECT * FROM test;")
	print(f"{len(rows)} row(s)")
	print(" ".join(fields.keys()))
	for r in rows:
		print(*r, sep=", ")
	db.close()
