# lidar system filter
#
# Copyright (c) 2023 ZKM | Hertz-Lab
# Dan Wilcox <dan.wilcox@zkm.de>
#
# BSD Simplified License.
# For information on usage and redistribution, and for a DISCLAIMER OF ALL
# WARRANTIES, see the file, "LICENSE.txt," in this distribution.
#
# This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
# Museum“ generously funded by the German Federal Cultural Foundation.
#
# ----- input
#
# json dict
#
# example lidar system frame message:
# {
#     "objects":[
#         {"object":{"type":"move","id":5,"x":32.810,"y":10.910,"size":0.300}},
#         {"object":{"type":"move","id":1692,"x":4.550,"y":16.750,"size":0.430}},
#         {"object":{"type":"move","id":1771,"x":26.100,"y":18.340,"size":0.270}},
#         {"object":{"type":"move","id":1772,"x":24.070,"y":18.320,"size":0.630}},
#         {"object":{"type":"move","id":1783,"x":-4.050,"y":8.050,"size":0.570}},
#         {"object":{"type":"move","id":1784,"x":1.140,"y":7.430,"size":0.300}}
#     ]
# }
# or single object messages with end of frame count:
# {"object":{"type":"move","id":2527,"x":12.000,"y":-12.330,"size":0.720}}
# {"count":1}
#
# object keys:
# * type: event type "enter", "move", "leave"
# * id: unique id (per-run)
# * x, y: position in meters, relative to current origin
# * size: size in meters, relative to tracked space w & h
#
# ----- output
#
# OSC-style json arrays
#
# person: id x y size type
# end of frame count: count
#
# normalized coordindates 0-1
#

import sys, json
from dataclasses import dataclass
import util

TOPIC = "lidar/telemetry"
NAME = "lidar"

# setup from a config dict with following keys
# * world: [w, h, x, y] world coordinate system in meteers
def setup(config):
    if "world" in config:
        world.set(config["world"])

# parse frame and send object(s)
def filter(client, data):
    if not isinstance(data, dict):
        print(f"lidar filter warning: payload not a dict", file=sys.stderr)
        return
    if "object" in data: # single object per message
        send_object(client, data)
    elif "objects" in data: # multi-object frame
        send_objects(client, data["objects"])
        send_count(client, len(data["objects"]))
    elif "count" in data: # single-object message end of frame count
        send_count(client, data["count"])
    else:
        print(f"lidar filter warning: no objects found", file=sys.stderr)

##### world

# world coordinate system in meters
@dataclass
class World:
    x: float = 0
    y: float = 0
    w: float = 100
    h: float = 100
    
    # set world coordinate system in meters: [x, y, w, h]
    # x,y: origin, w/h: size
    def set(self, world):
        # TODO: check ranges?
        self.x = world[0]
        self.y = world[1]
        self.w = world[2]
        self.h = world[3]

    # normalize a given x,y pos in world coordinates
    def normalize(self, x, y):
        return (
            util.clamp_value(util.map_value(x, self.x - self.w, self.w - self.x, 0, 1), 0, 1),
            util.clamp_value(util.map_value(y, self.y - self.h, self.h - self.y, 0, 1), 0, 1)
        )

# default instance
world = World()

##### sending

_TOPIC_PERSON = "lidar/person"
_TOPIC_COUNT = "lidar/count"
_SEND_OBJ = False

# filter object into person
def _filter_object(obj):
    try:
        x, y = world.normalize(obj["x"], obj["y"])
        if _SEND_OBJ:
            return {"id": obj["id"], "x": x, "y": y, "size": obj["size"], "type": obj["type"]}
        else:
            return [obj["id"], x, y, obj["size"], obj["type"]]
    except Exception as exc:
        print(f"lidar filter error: {exc}", file=sys.stderr)
    return None

# send object count
def send_count(client, count):
    try:
        if _SEND_OBJ:
            payload = {"count": count}
        else:
            payload = count
        client.publish(_TOPIC_COUNT, payload=json.dumps(payload), qos=0, retain=False)
        return True
    except Exception as exc:
        print(f"lidar send error: {exc}", file=sys.stderr)
    return False

# send an individual object
def send_object(client, obj):
    try:
        payload = _filter_object(obj["object"])
        if payload:
            client.publish(_TOPIC_PERSON, payload=json.dumps(payload), qos=0, retain=False)
            return True
    except Exception as exc:
        print(f"lidar send error: {exc}", file=sys.stderr)
    return False

# send an array fo objects
def send_objects(client, objs):
    for obj in objs:
        send_object(client, obj)
