// Copyright (c) 2023 ZKM | Hertz-Lab
// Dan Wilcox <dan.wilcox@zkm.de>
//
// BSD Simplified License.
// For information on usage and redistribution, and for a DISCLAIMER OF ALL
// WARRANTIES, see the file, "LICENSE.txt," in this distribution.
//
// This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
// Museum“ generously funded by the German Federal Cultural Foundation.

import Widget from "./Widget.js"

// tracked lidar person
export class Person {

  // create from JSON array [id x y size type], ex: [1, 0.2, 0.5, 2.1, "enter"] 
  static create(obj) {
    return new Person(obj[0], obj[1], obj[2], obj[3])
  }

  // create with unique id, normalized x y position 0-1, and relative size
  constructor(id, x, y, size) {
    this.id = id
    this.x = x
    this.y = y
    this.size = size
  }

  // set from JSON array [id x y size type]
  set(obj) {
    this.id = obj[0]
    this.x = obj[1]
    this.y = obj[2]
    this.size = obj[3]
  }
}

// lidar widget: displays current tracked people
export class Lidar extends Widget {

  constructor(args) {
  	args.unit = "visitors"
    super(args)
    this.data = new Map() // current people by unique id key
    this.scale.r = d3.scaleLinear()
          .domain([0, 1]) // data min max
          .range([8, 24]) // pixel min max
          .clamp(true)
  }

  // setup div with format:
  // <div id="lidar" class="widget widget-lidar">
  //     <svg id="lidar-svg" class="widget-svg" viewBox="0 0 W H"
  // preserveAspectRatio="xMidYMid meet"></svg>
  //     <div class="widget-title">LIDAR</div>
  //     <div class="widget-meta"><span class="widget-value">0</span> visitors</div>
  //     <div class="widget-controls">
  //         <button type="button">clear</button>
  //     </div>
  // </div>
  setup(w, h) {
    let self = this
    super.setup(w, h)
    
    this.div.classed("widget-lidar", true)
    this.div.selectAll(".widget-svg")
      .attr("font-size", "smaller")

    this.div.selectAll(".widget-controls")
      .append("button")
        .attr("type", "button")
        .on("click", function() {self.clear()})
        .text("clear")
  }

  // receive JSON message as a person
  receive(obj) {
    const id = obj[0]
    const event = obj[4]
    if(event == "leave") {
      this.data.delete(id)
      console.log("Lidar: removed " + id)
    }
    else { // enter or move
      let p = this.data.get(id)
      if(p === undefined) { // new
        this.data.set(id, Person.create(obj))
        console.log("Lidar: added " + id)
      }
      else { // exists
        p.set(obj)
      }
    }
  }

  // update svg from current data
  update() {
    let self = this

    // grouped circle and text
    // ref: https://www.d3indepth.com/transitions/
    // ref: https://stackoverflow.com/a/47401796/2146055
    this.svg.selectAll("g")
      .data(Array.from(this.data.values()))
      .join(
        function(enter) {
            let g = enter.append("g")
            .attr("transform", function(d) {
              return "translate(" + self.scale.x(d.x) + "," + self.scale.y(d.y) + ")"
            }.bind(self))
            .attr("opacity", 0)
            g.append("circle")
              .attr("r", function(d) {
                return self.scale.r(d.size)
              }.bind(self))
              .attr("fill", "gray")//d3.rgb(255, 0, 255))
              .attr("opacity", 0.5)
            g.append("text")
              .text(function(d) {return d.id})
              .attr("text-anchor", "middle")
              .attr("dominant-baseline", "middle")
              .attr("opacity", 0.75)
        return g
      },
      function(update) {
        return update
      },
      function(exit) {
        return exit
          .transition()
          .attr("opacity", 0)
          .remove()
      })
      .transition()
      .delay(250)
      .ease(d3.easeSinInOut)
      .attr("transform", function(d) {
          return "translate(" + self.scale.x(d.x) + "," + self.scale.y(d.y) + ")"
      }.bind(self))
      .attr("opacity", 1)
      .selectAll("circle")
        .attr("r", function(d) {
          return self.scale.r(d.size)
        }.bind(self))

    this.setValue(this.data.size)
  }

  clear() {
    this.data.clear()
    this.update()
  }

  // create random people
  randomize(count) {
    for(let i = 0; i < count; ++i) {
      let obj = [
        i,
        randomFloat(1),
        randomFloat(1),
        20 + randomInt(180),
        "move"  
      ]
      console.log(obj)
      let p = this.data.get(i)
      if(p === undefined) { // new
        this.data.set(i, Person.create(obj))
      }
      else {
        p.set(obj)
      }
    }
  }

}

export default Lidar
