// Copyright (c) 2023 ZKM | Hertz-Lab
// Dan Wilcox <dan.wilcox@zkm.de>
//
// BSD Simplified License.
// For information on usage and redistribution, and for a DISCLAIMER OF ALL
// WARRANTIES, see the file, "LICENSE.txt," in this distribution.
//
// This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
// Museum“ generously funded by the German Federal Cultural Foundation.

// D3 dashboard widget baseclass 
export class Widget {

  // create widget with required div id and svg virtual pixel size 
  // {id: "", size: [w, h]}
  // optional args:
  //  * title: display title string
  //  * unit: meta unit strings displayed after value
  //  * domain: {x: [min, max], y:[min, max]} input value ranges for each axis
  constructor(args) {
    this.id = args.id    // container div id, required
    this.div = null      // container div
    this.svg = null      // widget svg
    this.data = 0        // underlying data
    this.scale = {       // data scalers
      x: d3.scaleLinear()
          .domain([0, 1]) // data min max
          .range([0, 1])  // pixel min max
          .clamp(true),
      y: d3.scaleLinear()
          .domain([0, 1]) // data min max
          .range([0, 1])  // pixel min max
          .clamp(true)
    }
    this.title = (args.title === undefined ? "" : args.title)
    this.unit = (args.unit === undefined ? "" : args.unit)
    if(args.domain) {
    	if(args.domain.x) {this.scale.x.domain(args.domain.x)}
    	if(args.domain.y) {this.scale.y.domain(args.domain.y)}
    }
    if(args.size) {
      this.setup(args.size[0], args.size[1])
    }
  }

  // setup div id with basic format:
  // <div id="NAME" class="widget">
  //     <svg id="NAME-svg" class="widget-svg" viewBox="0 0 400 300"
  // preserveAspectRatio="xMidYMid meet"></svg>
  //     <div class="widget-title">TITLE</div>
  //     <div class="widget-meta"><span class="widget-value"></span></div>
  //     <div class="widget-controls"></div>
  // </div>
  setup(w, h) {
    // container
    this.div = d3.select("#" + this.id)
      .attr("class", "widget")
    
    // svg
    this.svg = this.div.append("svg")
      .attr("id", this.id + "-svg")
      .attr("class", "widget-svg")
      // .attr("viewBox", `0 0 1 1`)
      .attr("preserveAspectRatio", "xMidYMid meet")
    
    // title
    this.div.append("div")
      .attr("class", "widget-title")
      .text(this.title)
    
    // meta
    let valuehtml = "<span class=\"widget-value\"></span>"
    if(this.unit) {valuehtml += " "+this.unit}
    this.div.append("div")
      .attr("class", "widget-meta")
      .html(valuehtml)

    // controls
    this.div.append("div")
      .attr("class", "widget-controls")

    this.setSize(w, h)
  }

  // set virtual svg pixel size
  setSize(w, h) {
    this.scale.x.range([0, w])
    this.scale.y.range([0, h])
    this.svg.attr("viewBox", `0 0 ${w} ${h}`)
  }

  // set meta value display
  setValue(v) {
  	this.div.selectAll(".widget-value").text(v)
  }

  // update svg from current data
  update() {}

  // receive json object data, implement in subclass
  receive(data) {}

  // clear data
  clear() {
  	this.data = 0
  }
}

export default Widget
