// Copyright (c) 2023 ZKM | Hertz-Lab
// Dan Wilcox <dan.wilcox@zkm.de>
//
// BSD Simplified License.
// For information on usage and redistribution, and for a DISCLAIMER OF ALL
// WARRANTIES, see the file, "LICENSE.txt," in this distribution.
//
// This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
// Museum“ generously funded by the German Federal Cultural Foundation.
//
// References:
// * https://github.com/eclipse/paho.mqtt.javascript
// * https://www.d3indepth.com/selections/
// * https://www.d3indepth.com/shapes/
// * https://www.d3indepth.com/scales/

import {Widget} from "./Widget.js"
import Lidar from "./Lidar.js"
import Graph from "./Graph.js"

let status = {
  id: document.getElementById("status")
}

// ----- mqtt -----

// use localhost when only connecting internally
let host = "localhost"
let port = 9001
let topics = ["lidar/person", "lidar/count", "test/temperature", "test/humidity"]
let mqtt_id = "dashboard" // note: set to "" for server-generated id
let client = null // mqtt websocket

function mqtt_setup() {
  client = new Paho.Client(host, port, mqtt_id)
  client.onConnectionLost = (response) => {
    if(response.errorCode !== 0) {
      console.log("mqtt: disconnected: " + response.errorMessage)
    }
    status.id.textContent = "disconnected"
  }
  client.onMessageArrived = (message) => {
    //console.log("mqtt recv: ", message.destinationName, " ", message.payloadString)
    let data = JSON.parse(message.payloadString)
    if(data) {
      messageReceived(message.destinationName, data)
    }
  }
}

function mqtt_connect() {
  client.connect({
    onSuccess: () => {
      console.log("mqtt: connected")
      for(let t of topics) {
        client.subscribe(t)
      }
      status.id.textContent = "connected"
    },
    onFailure: (response) => {
      console.error("mqtt: connect failed: " + response.errorMessage)
      status.id.textContent = "connect failed"
    },
    cleanSession: true,
    reconnect: true
  })
}

function mqtt_disconnect() {
  if(client.isConnected()) {
    client.disconnect()
  }
}

// ----- main -----

let lidar = null
let temp = null
let humid = null

// parse url vars
let vars = getURLVars()
if("host" in vars) {
  host = vars.host
  console.log(`var host ${host}`)
}
if("port" in vars) {
  let n = parseInt(vars.port)
  if(n >= 1024) {
    port = n
    console.log(`var port ${port}`)
  }
}
if("id" in vars) {
  host = vars.host
  console.log(`var id ${host}`)
}
vars = null

// setup
window.addEventListener("load", () => {
  lidar = new Lidar({id: "lidar", title: "LIDAR", size: [640, 480]})
  temp = new Graph({id: "temp", title: "Temperature", unit: "°C", size: [400, 300], domain: {y: [40, -20]}, maxlen: 20})
  humid = new Graph({id: "humid", title: "Humidity", unit: "%", size: [400, 300], domain: {y: [100, 0]}})
  temp.set([0, 0, -10, -5, 1, 5, 10, 20, 15, 21])
  humid.set([0, 10, 20, 30, 40, 50, 60, 70, 80, 90])
  mqtt_setup()
  mqtt_connect()
})

// websocket message callback
// data sent as string & not JSON?
function messageReceived(topic, data) {
  if(topic == "lidar/person") {
    if(data == "") {return}
    if(data instanceof Blob) {
      data.text().then(text => lidar.receive(JSON.parse(text)))
    }
    else {
      if(typeof data === "string") {data = JSON.parse(data)}
      lidar.receive(data)
    }
  }
  else if(topic == "lidar/count") {
    if(typeof data === "string") {data = JSON.parse(data)}
    if(data == 0) {
      lidar.clear()
    }
    else {
      lidar.update()
    }
  }
  else if(topic == "test/temperature") {
    if(typeof data === "string") {data = JSON.parse(data)}
    temp.receive(data)
    temp.update()
  }
  else if(topic == "test/humidity") {
    if(typeof data === "string") {data = JSON.parse(data)}
    humid.receive(data)
    humid.update()
  }
}

function reconnect() {
  mqtt_disconnect()
  mqtt_connect()
}

// ----- util -----

// returns URL variables as key/value pairs:
// ?foo=bar&baz=123 -> {foo: "bar", baz: 123}
function getURLVars() {
  let vars = {}
  let parts = window.location.href.replace(/[?&]+([^=&]+)=([^&]*)/gi, function(m, key, value) {
    vars[key] = value
  })
  return vars
}

// returns string as a bool
function parseBool(s) {
  return (s === "true" || s === "1")
}

// returns a random float 0 - x
function randomFloat(x) {
  return Math.random() * x
}

// returns a random int 0 - x
function randomInt(x) {
  return Math.floor(Math.random() * x)
}
