/* p5.js template
   Dan Wilcox ZKM | Hertz-Lab 2022 */

// ref: https://developer.mozilla.org/en-US/docs/Web/API/WebSocket

let status = {
  id: document.getElementById("status")
}
let visitors = {
  id: document.getElementById("visitors")
}

// ----- mqtt -----

// use localhost when only connecting internally
let host = "localhost"
let port = 9001
let topics = ["lidar/person", "lidar/count"]
let id = "wsclient" // note: set to "" for server-generated id

// ----- Blob -----

class Person {

  // create from JSON array [id x y size type], ex: [1, 100, 100, size, enter] 
  static create(obj) {
    return new Person(obj[0], obj[1], obj[2], obj[3])
  }

  constructor(id, x, y, size) {
    this.id = id
    this.pos = new p5.Vector(x, y)
    this.size = size
  }

  set(obj) {
    this.id = obj[0]
    this.pos.x = obj[1]
    this.pos.y = obj[2]
    this.size = obj[3]
  }

  draw() {
    let x = map(this.pos.x, 0, 1, 0, width)
    let y = map(this.pos.y, 0, 1, 0, height)
    let s = map(this.size, 0, 1, 10, 50) 
    noStroke()
    fill(max(min(map(this.size, 0, 1, 20, 200), 20), 200))
    circle(x, y, s)
  }

  drawId() {
    let x = map(this.pos.x, 0, 1, 0, width)
    let y = map(this.pos.y, 0, 1, 0, height)
    noStroke()
    fill(0)
    textAlign(CENTER, CENTER)
    text(this.id, x, y)
  }

}

// ----- p5 sketch -----

let people = new Map()
let client = null

function preload() {

}

function setup() {

  // fill screen
  canvas = createCanvas(windowWidth, windowHeight)
  canvas.parent("sketch")
  frameRate(30)
  smooth()

  // mqtt websocket
  client = new Paho.Client(host, port, id)
  client.onConnectionLost = (response) => {
    if(response.errorCode !== 0) {
      console.log("mqtt: disconnected: " + response.errorMessage)
    }
    status.id.textContent = "disconnected"
  }
  client.onMessageArrived = (message) => {
    //console.log("mqtt recv: ", message.destinationName, " ", message.payloadString)
    let data = JSON.parse(message.payloadString)
    if(data) {
      messageReceived(message.destinationName, data)
    }
  }
  client.connect({
    onSuccess: () => {
      console.log("mqtt: connected")
      for(let t of topics) {
        client.subscribe(t)
      }
      status.id.textContent = "connected"
    },
    onFailure: (response) => {
      console.error("mqtt: connect failed: " + response.errorMessage)
      status.id.textContent = "connect failed"
    },
    // optional authentication
    //userName: "test",
    //password: "test123"
  })
}

function draw() {
  background(255)
  for(const[id, p] of people) {
    p.draw()
  }
  for(const[id, p] of people) {
    p.drawId()
  }
}

function mouseMoved() {

}

function mousePressed() {

}

function mouseDragged() {

}

function mouseReleased() {

}

function keyPressed() {

}

function keyReleased() {

}

function windowResized() {
  resizeCanvas(windowWidth, windowHeight)
}

// websocket message callback
// data sent as string & not JSON?
function messageReceived(topic, data) {
  if(topic == "lidar/person") {
    if(data == "") {return}
    if(data instanceof Blob) {
      data.text().then(text => receivePerson(JSON.parse(text)))
    }
    else {
      if(typeof data === "string") {data = JSON.parse(data)}
      receivePerson(data)
    }
  }
  else if(topic == "lidar/count") {
    visitors.id.textContent = data
  }
}

function receivePerson(data) {
  // if(data.length() < 5 && data[0] === null || data[1] === null || data["y"]  === null ||
  //    data["size"] === null || data["type"] === null) {return}
  const id = data[0]
  const event = data[4]
  if(event == "leave") {
    people.delete(id)
    console.log("removed " + id)
  }
  else { // enter or move
    let p = people.get(id)
    if(p === undefined) { // new
      people.set(id, Person.create(data))
      console.log("added " + id)
    }
    else { // exists
      p.set(data)
    }
  }
}

// ----- util -----

let util = {

debug: true,

// set debug state
setDebug: function(enable) {
  util.debug = enable
},

// only print if debug is set
debugPrint: function(msg) {
  if(util.debug) {
    console.log(msg)
  }
},

// returns URL variables as key/value pairs:
// ?foo=bar&baz=123 -> {foo: "bar", baz: 123}
getURLVars: function() {
  let vars = {}
  let parts = window.location.href.replace(/[?&]+([^=&]+)=([^&]*)/gi, function(m, key, value) {
    vars[key] = value
  })
  return vars
},

// returns string as a bool
parseBool: function(s) {
  return (s === "true" || s === "1")
}

} // util

// ----- main -----

// parse url vars
let vars = util.getURLVars()
if("debug" in vars) {
  util.setDebug(util.parseBool(vars.debug))
  console.log(`var debug ${util.debug}`)
}
if("host" in vars) {
  host = vars.host
  console.log(`var host ${host}`)
}
if("port" in vars) {
  let n = parseInt(vars.port)
  if( n >= 1024) {
    port = n
    console.log(`var port ${port}`)
  }
}
vars = null
