// refs:
// * https://github.com/eclipse/paho.mqtt.javascript
// * https://www.d3indepth.com/selections/
// * https://www.d3indepth.com/shapes/
// * https://www.d3indepth.com/scales/

let status = {
  id: document.getElementById("status")
}

// ----- mqtt -----

// use localhost when only connecting internally
let host = "localhost"
let port = 9001
let topics = ["lidar/person", "lidar/count"]
let mqtt_id = "wsclient" // note: set to "" for server-generated id
let client = null // mqtt websocket

function mqtt_setup() {
  client = new Paho.Client(host, port, "lidar-web")
  client.onConnectionLost = (response) => {
    if(response.errorCode !== 0) {
      console.log("mqtt: disconnected: " + response.errorMessage)
    }
    status.id.textContent = "disconnected"
  }
  client.onMessageArrived = (message) => {
    //console.log("mqtt recv: ", message.destinationName, " ", message.payloadString)
    let data = JSON.parse(message.payloadString)
    if(data) {
      messageReceived(message.destinationName, data)
    }
  }
}

function mqtt_connect() {
  client.connect({
    onSuccess: () => {
      console.log("mqtt: connected")
      for(let t of topics) {
        client.subscribe(t)
      }
      status.id.textContent = "connected"
    },
    onFailure: (response) => {
      console.error("mqtt: connect failed: " + response.errorMessage)
      status.id.textContent = "connect failed"
    },
    cleanSession: true,
    reconnect: true
  })
}

function mqtt_disconnect() {
    client.disconnect()
}

// ----- d3 -----

class Widget {

  // {id: "", title: "", size: [w, h]}
  constructor(args) {
    this.id = args.id    // container div id
    this.data = []  // underlying data
    this.scale = {  // data scalers
      x: d3.scaleLinear()
          .domain([0, 1]) // data min max
          .range([0, 1])  // pixel min max
          .clamp(true),
      y: d3.scaleLinear()
          .domain([0, 1]) // data min max
          .range([0, 1])  // pixel min max
          .clamp(true)
    }
    this.title = (args.title === null ? "" : args.title)
    if(args.size) {
      this.setup(args.size[0], args.size[1])
    }
  }

  // <div id="NAME" class="widget">
  //     <svg id="NAME-svg" class="widget-svg" viewBox="0 0 800 600"
  // preserveAspectRatio="xMidYMid meet"></svg>
  //     <div class="widget-title">TITLE</div>
  //     <div class="widget-meta"></div>
  //     <div class="widget-controls"></div>
  // </div>
  setup(w, h) {
    // container
    let div = d3.select("#" + this.id)
      .attr("class", "widget")
    
    // svg
    div.append("svg")
      .attr("id", this.id + "-svg")
      .attr("class", "widget-svg")
      .attr("viewBox", `0 0 1 1`)
      .attr("preserveAspectRatio", "xMidYMid meet")
    
    // title
    div.append("div")
      .attr("class", "widget-title")
      .text(this.title)
    
    // meta
    div.append("div")
      .attr("class", "widget-meta")

    // controls
    div.append("div")
      .attr("class", "widget-controls")

    this.setSize(w, h)
  }

  setSize(w, h) {
    this.scale.x.range([0, w])
    this.scale.y.range([0, h])
    d3.select("#" + this.id + "-svg")
      .attr("viewBox", `0 0 ${w} ${h}`)
  }

  update() {

  }

  set(data) {

  }

  receive(data) {

  }
}

class Person {

  // create from JSON array [id x y size type], ex: [1, 100, 100, size, enter] 
  static create(obj) {
    return new Person(obj[0], obj[1], obj[2], obj[3])
  }

  constructor(id, x, y, size) {
    this.id = id,
    this.x = x
    this.y = y
    this.size = size
  }

  set(obj) {
    this.id = obj[0]
    this.x = obj[1]
    this.y = obj[2]
    this.size = obj[3]
  }
}

class Lidar extends Widget {

  constructor(args) {
    super(args)
    this.data = new Map() // current people by unique id key
    this.scale.r = d3.scaleLinear()
          .domain([0, 1]) // data min max
          .range([8, 24]) // pixel min max
          .clamp(true)
  }

  // <div id="lidar" class="widget">
  //     <svg id="lidar-svg" class="widget-svg" viewBox="0 0 W H"
  // preserveAspectRatio="xMidYMid meet"></svg>
  //     <div class="widget-title">LIDAR</div>
  //     <div class="widget-meta"><span id="lidar-visitors">0</span> visitors</div>
  //     <div class="widget-controls">
  //         <button type="button" onclick="clearLidar()">clear</button>
  //     </div>
  // </div>
  setup(w, h) {
    let self = this
    super.setup(w, h)
    
    let div = d3.select("#"+this.id)
    div.selectAll(".widget-svg")
      .attr("font-size", "smaller")

    div.selectAll(".widget-meta")
      .html("<span id=\"lidar-visitors\">0</span> visitors")

    div.selectAll(".widget-controls")
      .append("button")
        .attr("type", "button")
        .on("click", function() {self.clear()})
        .text("clear")
  }

  // receive JSON message as a person
  receive(obj) {
    const id = obj[0]
    const event = obj[4]
    if(event == "leave") {
      this.data.delete(id)
      console.log("Lidar: removed " + id)
    }
    else { // enter or move
      let p = this.data.get(id)
      if(p === undefined) { // new
        this.data.set(id, Person.create(obj))
        console.log("Lidar: added " + id)
      }
      else { // exists
        p.set(obj)
      }
    }
  }

  // set virtual svg pixel size
  setSize(w, h) {
    this.scale.x.range([0, w])
    this.scale.y.range([0, h])
    d3.select("#" + this.id + "-svg")
      .attr("viewBox", `0 0 ${w} ${h}`)
  }

  // update svg from current data
  update() {
    let self = this

    // simple circle position
    // d3.select("#"+this.id+"-svg")
    //   .selectAll("circle")
    //   .data(Array.from(this.data.values()))
    //   .join("circle")
    //     .attr('fill', d3.rgb(255, 0, 255))
    //     .attr("opacity", 0.5)
    //     //.transition()
    //     .attr("cx", function(d) {
    //       return lidar.scale.x(d.x)
    //     })
    //     .attr("cy", function(d) {
    //       return lidar.scale.y(d.y)
    //     })
    //     .attr("r", function(d) {
    //       return lidar.scale.r(d.size)
    //     })


    // circle position with transitions
    // d3.select("#"+this.id+"-svg")
    //   .selectAll("circle")
    //   .data(Array.from(this.data.values()))
    //   .join(
    //     function(enter) {
    //       return enter.append("circle")
    //         .attr("cx", function(d) {
    //           return lidar.scale.x(d.x)
    //         })
    //         .attr("cy", function(d) {
    //           return lidar.scale.y(d.y)
    //         })
    //         .attr("r", function(d) {
    //           return lidar.scale.r(d.size)
    //         })
    //         .attr("fill", d3.rgb(255, 0, 255))
    //         .attr("opacity", 0)
    //     },
    //     function(update) {
    //       return update
    //     },
    //     function(exit) {
    //       return exit
    //         .transition()
    //         .attr("opacity", 0)
    //         .remove()
    //     }
    //   )
    //   .transition()
    //   .duration(250)
    //   .ease(d3.easeSinInOut)
    //   .attr("cx", function(d) {
    //     return lidar.scale.x(d.x)
    //   })
    //   .attr("cy", function(d) {
    //     return lidar.scale.y(d.y)
    //   })
    //   .attr("r", function(d) {
    //     return lidar.scale.r(d.size)
    //   })
    //   .attr("opacity", 0.5)

    // grouped circle and text
    // ref: https://www.d3indepth.com/transitions/
    // ref: https://stackoverflow.com/a/47401796/2146055
    d3.select("#"+this.id+"-svg")
      .selectAll("g")
      .data(Array.from(this.data.values()))
      .join(
        function(enter) {
            let g = enter.append("g")
            .attr("transform", function(d) {
              return "translate(" + self.scale.x(d.x) + "," + self.scale.y(d.y) + ")"
            }.bind(self))
            .attr("opacity", 0)
            g.append("circle")
              .attr("r", function(d) {
                return self.scale.r(d.size)
              }.bind(self))
              .attr("fill", "gray")//d3.rgb(255, 0, 255))
              .attr("opacity", 0.5)
            g.append("text")
              .text(function(d) {return d.id})
              .attr("text-anchor", "middle")
              .attr("dominant-baseline", "middle")
              .attr("opacity", 0.75)
        return g
      },
      function(update) {
        return update
      },
      function(exit) {
        return exit
          .transition()
          .attr("opacity", 0)
          .remove()
      })
      .transition()
      .delay(250)
      .ease(d3.easeSinInOut)
      .attr("transform", function(d) {
          return "translate(" + self.scale.x(d.x) + "," + self.scale.y(d.y) + ")"
      }.bind(self))
      .attr("opacity", 1)
      .selectAll("circle")
        .attr("r", function(d) {
          return self.scale.r(d.size)
        }.bind(self))

    d3.select("#"+this.id+"-visitors").text(this.data.size)
  }

  clear() {
    this.data.clear()
    this.update()
  }

  // create random people
  randomize(count) {
    for(let i = 0; i < count; ++i) {
      let obj = [
        i,
        randomFloat(1),
        randomFloat(1),
        20 + randomInt(180),
        "move"  
      ]
      console.log(obj)
      let p = this.data.get(i)
      if(p === undefined) { // new
        this.data.set(i, Person.create(obj))
      }
      else {
        p.set(obj)
      }
    }
  }

}

// ----- main -----

let lidar = null

// parse url vars
let vars = getURLVars()
if("host" in vars) {
  host = vars.host
  console.log(`var host ${host}`)
}
if("port" in vars) {
  let n = parseInt(vars.port)
  if(n >= 1024) {
    port = n
    console.log(`var port ${port}`)
  }
}
if("id" in vars) {
  host = vars.host
  console.log(`var id ${host}`)
}
vars = null

// setup
window.addEventListener("load", () => {
  lidar = new Lidar({id: "lidar", title: "LIDAR", size:[800, 600]})
  mqtt_setup()
  mqtt_connect()
})

// websocket message callback
// data sent as string & not JSON?
function messageReceived(topic, data) {
  if(topic == "lidar/person") {
    if(data == "") {return}
    if(data instanceof Blob) {
      data.text().then(text => lidar.receive(JSON.parse(text)))
    }
    else {
      if(typeof data === "string") {data = JSON.parse(data)}
      lidar.receive(data)
    }
  }
  else if(topic == "lidar/count") {
    if(typeof data === "string") {data = JSON.parse(data)}
    if(data == 0) {
      lidar.clear()
    }
    else {
      lidar.update()
    }
  }
}

function reconnect() {
  mqtt_disconnect()
  mqtt_connect()
}

// ----- util -----

// returns URL variables as key/value pairs:
// ?foo=bar&baz=123 -> {foo: "bar", baz: 123}
function getURLVars() {
  let vars = {}
  let parts = window.location.href.replace(/[?&]+([^=&]+)=([^&]*)/gi, function(m, key, value) {
    vars[key] = value
  })
  return vars
}

// returns string as a bool
function parseBool(s) {
  return (s === "true" || s === "1")
}

// returns a random float 0 - x
function randomFloat(x) {
  return Math.random() * x
}

// returns a random int 0 - x
function randomInt(x) {
  return Math.floor(Math.random() * x)
}
