#! /usr/bin/env python3
#
# Copyright (c) 2023 ZKM | Hertz-Lab
# Dan Wilcox <dan.wilcox@zkm.de>
#
# BSD Simplified License.
# For information on usage and redistribution, and for a DISCLAIMER OF ALL
# WARRANTIES, see the file, "LICENSE.txt," in this distribution.
#
# This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
# Museum“ generously funded by the German Federal Cultural Foundation.
#
# References:
# * https://github.com/eclipse/paho.mqtt.python
# * https://github.com/attwad/python-osc

import asyncio, argparse, sys, json, time
import paho.mqtt.client as mqtt

import config
import sensor

mqtt_host = "localhost"
mqtt_port = 1883
mqtt_user = ""
mqtt_pass = ""

verbose = False

##### parser

parser = argparse.ArgumentParser(description="(fake) sensors -> MQTT")
parser.add_argument(
    "host", type=str, nargs="?", metavar="HOST",
    default="", help="MQTT server host name or IP address, optionally append port HOST:PORT")
parser.add_argument(
    "--user", action="store", dest="user",
    default="", help="MQTT server user name")
parser.add_argument(
    "--pass", action="store", dest="password",
    default="", help="MQTT server user password")
parser.add_argument(
    "-v", "--verbose", action="store_true", dest="verbose",
    help="enable verbose printing")

##### mqtt

# mqtt connection callback
def mqtt_on_connect(client, userdata, flags, rc):
    if rc == 0:
        if verbose: print(f"mqtt: connected")
    else: print(f"mqtt: {mqtt.error_string(rc)}")

# stop osc receiver on unplanned disconnect
def mqtt_on_disconnect(client, userdata, rc):
    if rc != 0:
        print(f"mqtt: {mqtt.error_string(rc)}")
        asyncio.get_running_loop().stop()

##### main

# parse config
args = parser.parse_args()
if args.host == "":
        print("MQTT HOST required", file=sys.stderr)
        sys.exit(1)
if ":" in args.host: # parse optional HOST:PORT
    s = args.host.split(":")
    try: # strip last component
        mqtt_port = int(s[-1])
    except:
        print(f"invalid port {s[-1]} in MQTT HOST {args.host}", file=sys.stderr)
        sys.exit(1)
    # join everything else, in case of domain ie. http://
    mqtt_host = ":".join(s[:-1])
else:
    mqtt_host = args.host
if args.user != "":
    if args.password == "":
        print(f"password required for MQTT server user {args.user}", file=sys.stderr)
        sys.exit(1)
    else:
        mqtt_user = args.user
        mqtt_pass = args.password
verbose = args.verbose
args = None

# mqtt client
client = mqtt.Client()
client.on_connect = mqtt_on_connect
client.on_disconnect = mqtt_on_disconnect

# run sensors (coroutines)
async def main():
    print(f"sensors -> mqtt {mqtt_host}:{mqtt_port}")
    try:
        if mqtt_user:
            client.username_pw_set(username=mqtt_user, password=mqtt_pass)
        client.connect(mqtt_host, mqtt_port, 60)
    except Exception as exc:
        print(exc, file=sys.stderr)
        sys.exit(1)
    tasks = []
    for s in config.sensors:
        s.verbose = verbose
        tasks.append(sensor.create_task(s, client))
    await asyncio.gather(*tasks)
try:
    asyncio.run(main())
except KeyboardInterrupt:
    pass
finally:
    client.disconnect()
