# Copyright (c) 2023 ZKM | Hertz-Lab
# Dan Wilcox <dan.wilcox@zkm.de>
#
# BSD Simplified License.
# For information on usage and redistribution, and for a DISCLAIMER OF ALL
# WARRANTIES, see the file, "LICENSE.txt," in this distribution.
#
# This code has been developed at ZKM | Hertz-Lab as part of „The Intelligent
# Museum“ generously funded by the German Federal Cultural Foundation.
#
# References:
# * https://www.pythontutorial.net/python-concurrency/python-asyncio-create_task/

import asyncio, time, random, sys, json

# create async io task to run a sensor instance with mqtt client
def create_task(s, client):
    return asyncio.create_task(_run(s, client))

# sensor run callback with mqtt client
async def _run(s, client):
    s.client = client
    if s.verbose: s.print()
    while True:
        await asyncio.sleep(s.interval)
        s.update()
        if s.verbose: s.print()
        s.send()

# baseclass
class Sensor:

    # constructor with required mqtt topic
    # optional
    # * interval: update frequency in s
    # * client: set mqtt client on creation
    # * verbose: print values after update?
    def __init__(self, topic, interval=1, client=None, verbose=False):
        self.topic = topic
        self.interval = interval
        self.client = client # mqtt client

    # update value
    # implement in child
    def update(self):
        pass

    # create mqtt (json) payload
    # implement in child
    def payload(self):
        return ""

    # print topic and current value(s)
    # implement in child
    def print(self):
        print(f"{self.topic}")

    # send callback
    def send(self):
        if self.client:
            try:
                self.client.publish(self.topic, payload=json.dumps(self.payload()), qos=0, retain=False)
            except Exception as exc:
                print(f"{self.topic} {exc}", file=sys.stderr)

    # create async io task to run this sensor instance
    @staticmethod
    def create_task(s):
        return asyncio.create_task(Sensor._run(s))

    # sensor run callback
    @staticmethod
    async def _run(s):
        if s.verbose: s.print()
        while True:
            await asyncio.sleep(s.interval)
            s.update()
            if s.verbose: s.print()
            s.send()

# single value with random walk
class Value(Sensor):

    # constructor with required MQTT topic
    # optional
    # * interval: update frequency in s
    # * value: starting value
    # * step: how much value can change on each update
    # * range: [min, max] allowed value range
    def __init__(self, topic, interval=1, value=0, step=1, range=None):
        super().__init__(topic, interval)
        self.value = value
        self.step = step
        self.range = range # [min, max]

    def update(self):
        dv = random.choice([-self.step, self.step])
        if self.range:
            self.value = clamp(self.value + dv, self.range[0], self.range[1])
        else:
            self.value += dv

    def payload(self):
        return self.value

    def print(self):
        print(f"{self.topic} {self.value}")

# lidar tracked person
class LidarPerson(Sensor):

    # constructor with required MQTT topic and unqiue id (per-run)
    def __init__(self, topic, id):
        super().__init__(topic)
        self.id = id
        self.x = 0
        self.y = 0
        self.vx = 0
        self.vy = 0
        self.size = 0.5
        self.type = "enter"
        self.random()

    def update(self):
        ret = False
        if self.type == "enter":
            self.send()
            self.type = "move"
            return ret
        self.x += self.vx
        self.y += self.vy
        if self.x < 0 or self.x > 1 or self.y < 0 or self.y > 1:
            self.type = "leave"
            ret = True
        self.send()
        return ret

    def payload(self):
        return [self.id, clamp(self.x, 0, 1), clamp(self.y, 0, 1), self.size, self.type]

    def print(self):
        print(f"{self.topic} {self.id} {self.x} {self.y} {self.size} {self.type}")

    def random(self):
        self.x = random.random()
        self.y = random.random()
        self.vx = random.random() * random.choice([-0.02, -0.01, 0.01, 0.02]) + random.choice([-0.001, 0.001])
        self.vy = random.random() * random.choice([-0.02, -0.01, 0.01, 0.02]) + random.choice([-0.001, 0.001])
        self.size = random.randrange(1, 50) / 10

# aggregate Lidar system
class Lidar(Sensor):

    # constructor with required MQTT topic
    # optional
    # * interval: update frequency in s
    # * maxpeople: max allowed people to generate
    # * newchance: percentage change to add new person 0-1, ie. 0.3 = 30%
    def __init__(self, basetopic, interval=0.2, maxpeople=20, newchance=0.3):
        super().__init__(basetopic+"/count", interval)
        self.basetopic = basetopic
        self.people = []
        self.id = 0
        self.max = maxpeople
        self.newchance = newchance
        self.firstframe = True

    # add, update, or remove people
    def update(self):
        rem = []
        for i, p in enumerate(self.people):
            if p.update(): rem.append(i)
        for i in rem:
            try:
                self.people.pop(i)
            except Exception:
                # FIXME: not sure why this happens...
                pass
        # make sure first frame is empty so clients can clear themselves
        if self.firstframe:
            self.firstframe = False
            return
        # add new person?
        if len(self.people) < self.max and random.random() <= self.newchance:
            p = self.add()

    # add a new person
    def add(self):
        p = LidarPerson(topic=self.basetopic+"/person", id=self.id)
        p.client = self.client
        self.people.append(p)
        self.id += 1
        return p

    # end of frame count
    def payload(self):
        return str(len(self.people))

    def print(self):
        for p in self.people:
            p.print()
        print(f"{self.topic} {len(self.people)}")

# util

# clamp value v between range min and max
def clamp(v, rmin, rmax):
    return max(rmin, min(v, rmax))
